# Copyright (c) 2014 Thomas Heller
# Copyright (c) 2015 Andreas Schaefer
#
# Distributed under the Boost Software License, Version 1.0. (See accompanying
# file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

# include exported targets if not already defined
if (NOT TARGET hpx)
  include("${CMAKE_CURRENT_LIST_DIR}/HPXTargets.cmake")
endif()

# include HPX cmake utilities
include("${CMAKE_CURRENT_LIST_DIR}/HPXMacros.cmake")

get_filename_component(_hpx_root_dir "${CMAKE_CURRENT_LIST_DIR}/../../.." ABSOLUTE)

set(HPX_INCLUDE_DIRS "/usr/include;/usr/include;/usr/include")

set(HPX_LIBRARIES "general;hpx_init;general;hpx;atomic;dl;rt;/usr/lib64/libboost_filesystem.so;/usr/lib64/libboost_program_options.so;/usr/lib64/libboost_system.so;/usr/lib64/libpthread.so;/usr/lib64/libboost_filesystem.so;/usr/lib64/libboost_program_options.so;/usr/lib64/libboost_system.so;/usr/lib64/libtcmalloc_minimal.so;/usr/lib64/libhwloc.so")
set(HPX_LIBRARY_DIR "")

set(HPX_VERSION_STRING "1.3.0")
set(HPX_VERSION_MAJOR 1)
set(HPX_VERSION_MINOR 3)
set(HPX_VERSION_SUBMINOR 0)

set(HPX_PREFIX "${_hpx_root_dir}")
set(HPX_DEBUG_POSTFIX "d")
set(HPX_BUILD_TYPE "Release")

set(HPX_GIT_COMMIT "unknown" CACHE STRING "Revision of HPX from Git" FORCE)

set(HPX_C_COMPILER "/usr/bin/cc" CACHE STRING "C compiler for HPX" FORCE)
set(HPX_C_COMPILER_ID "GNU" CACHE STRING "C compiler id for HPX" FORCE)
set(HPX_C_COMPILER_VERSION "9.1.1" CACHE STRING "C compiler version for HPX" FORCE)

set(HPX_CXX_COMPILER "/usr/bin/c++" CACHE STRING "CXX compiler for HPX" FORCE)
set(HPX_CXX_COMPILER_ID "GNU" CACHE STRING "CXX compiler id for HPX" FORCE)
set(HPX_CXX_COMPILER_VERSION "9.1.1" CACHE STRING "CXX compiler version for HPX" FORCE)

set(HPX_Fortran_COMPILER "" CACHE STRING "Fortran compiler for HPX" FORCE)
set(HPX_Fortran_COMPILER_ID "" CACHE STRING "Fortran compiler id for HPX" FORCE)
set(HPX_Fortran_COMPILER_VERSION "" CACHE STRING "Fortran compiler version for HPX" FORCE)

set(HPX_BOOST_ROOT "")
set(HPX_BOOST_LIBRARIES "/usr/lib64/libboost_filesystem.so;/usr/lib64/libboost_program_options.so;/usr/lib64/libboost_system.so;/usr/lib64/libpthread.so;/usr/lib64/libboost_filesystem.so;/usr/lib64/libboost_program_options.so;/usr/lib64/libboost_system.so")

set(HPX_WITH_PSEUDO_DEPENDENCIES ON)
set(HPX_WITH_STATIC_LINKING OFF)
set(HPX_WITH_MALLOC_DEFAULT tcmalloc)
set(HPX_WITH_PARCELPORT_TCP ON)
set(HPX_WITH_PARCELPORT_MPI OFF)
set(HPX_WITH_APEX OFF)
if(MSVC)
  set(HPX_WITH_VCPKG )
endif()
set(HPX_WITH_DYNAMIC_HPX_MAIN ON)
set(HPX_WITH_DATAPAR )
set(HPX_WITH_DATAPAR_VC OFF)

if(NOT HPX_CMAKE_LOGLEVEL)
  set(HPX_CMAKE_LOGLEVEL "WARN")
endif()

set(HPX_LINK_LIBRARIES "general;hpx_wrap;")

enable_language(C)

hpx_check_compiler_compatibility()
hpx_check_boost_compatibility()
hpx_setup_allocator()

if(NOT DEFINED ${CMAKE_FIND_PACKAGE_NAME}_FOUND)
  set(${CMAKE_FIND_PACKAGE_NAME}_FOUND true)
endif()

# Set variables to enable cuda/nvcc or cuda/clang when using add_hpx_executable()
set(HPX_WITH_CUDA OFF)
set(HPX_WITH_CUDA_CLANG OFF)
set(HPX_CUDA_CLANG_FLAGS )

# Set variables used by nvcc
if(HPX_WITH_CUDA)
    find_package(CUDA REQUIRED)
    set(HPX_WITH_CUDA_CLANG OFF)
    set(CUDA_SEPARABLE_COMPILATION ON)
    set(CUDA_NVCC_FLAGS )
    set(HPX_CUDA_CLANG_FLAGS )
    # The following is a workaround to make sure that target_link_libraries in
    # cuda_add_executable/library is called with a link keyword, not with a
    # plain signature. From CMake 3.9 onwards this can be done officially with
    # the CUDA_LINK_LIBRARIES_KEYWORD. For older versions we can emulate this by
    # prepending a link keyword to CUDA_LIBRARIES.
    #
    # See
    # https://gitlab.kitware.com/cmake/cmake/commit/9f41bfd7b9e6e8a7545f741f872949d2ae801978
    # and https://gitlab.kitware.com/cmake/cmake/issues/16602 for more details.
    if(CMAKE_VERSION VERSION_LESS "3.9")
      set(CUDA_LIBRARIES PUBLIC ${CUDA_LIBRARIES})
      set(CUDA_cudadevrt_LIBRARY PUBLIC ${CUDA_cudadevrt_LIBRARY})
    else()
      set(CUDA_LINK_LIBRARIES_KEYWORD PUBLIC)
    endif()
    cuda_include_directories("/usr/include;/usr/include;/usr/include")
endif()

# Add HPX linker flag to CMAKE_EXE_LINKER_FLAGS
if(HPX_WITH_DYNAMIC_HPX_MAIN AND ("${CMAKE_SYSTEM_NAME}" STREQUAL "Linux"))
    set(HPX_LINKER_FLAGS "-Wl,-wrap=main")
endif()
